@@ -217,21 +217,34 @@ text_fuzzy_status_t;
 #include <string.h>
 #include <errno.h>
 #include <limits.h>
+#include <stdint.h>
 
 #include "config.h"
 #include "text-fuzzy.h"
 
-/* All of the following are automatically generated from
-   "edit-distance.h.tmpl" by "make-edit-distance-c.pl". */
+#ifdef HEADER
 
-#include "edit-distance-char-trans.h"
-#include "edit-distance-int-trans.h"
-#include "edit-distance-char.h"
-#include "edit-distance-int.h"
+/* Some people are not using GCC-compatible compilers. See, for
+   example, this test failure.
+   http://ppm4.activestate.com/sun4-solaris-64/5.12/1200/B/BK/BKB/Text-Fuzzy-0.11.d/log-20130328T025706.txt */
 
-/* The following starts off the header file. This is a tag which tells
-   "C::Maker" to start writing "text-fuzzy.h". */
+#ifdef __GNUC__
+#define STACKALLOCOK 1
+#else
+/* User's compiler cannot allocate on the stack. */
+#undef STACKALLOCOK
+#endif
 
+#define CALLOC(x,n,s) {						\
+	x = calloc (n, s);					\
+	if (! x) {						\
+	    fprintf (stderr, "%s:%d: calloc %d x %d failed.\n",	\
+		     __FILE__, __LINE__, n, s);			\
+	    return -1;						\
+	}							\
+    }
+
+#endif /* def HEADER */
 #ifdef HEADER
 
 //#define VERBOSE 1
@@ -285,12 +298,45 @@ text_fuzzy_string_t;
 
 typedef struct candidate candidate_t;
 
-struct candidate {
+struct candidate
+{
     int distance;
     int offset;
     candidate_t * next;
 };
 
+/* Dictionary for transposition character matches. The "a" in "adic_t"
+   stands for "ascii". */
+
+typedef struct
+{
+    int dic[UINT8_MAX];
+}
+adic_t;
+
+/* One item in dictionary for transposition integer matches. */
+
+typedef struct item
+{
+    /* The character. */
+    int key; 
+    /* The position of the character in the first string. */
+    unsigned int position;
+}
+idic_item_t;
+
+/* Dictionary of last-seen positions of characters for transposition
+   integer matches. */
+
+typedef struct
+{
+    /* Allocated by size of dictionary. */
+    idic_item_t * items;
+    /* Largest value. */
+    int max;
+}
+idic_t;
+
 /* The following structure contains one string plus additional
    paraphenalia used in searching for the string, for example the
    alphabet of the string. */
@@ -355,6 +401,15 @@ typedef struct text_fuzzy
 
     int offset;
 
+    /* ASCII match dictionary for transposition searches. */
+
+    adic_t adic;
+
+    /* Unicode match dictionary (integer) for transposition
+       searches. */
+
+    idic_t idic;
+
     /* Does the user want to use an alphabet filter? Default is yes,
        so this must be set to a non-zero value to switch off use. */
     unsigned int user_no_alphabet : 1;
@@ -385,6 +440,10 @@ typedef struct text_fuzzy
 
     /* Do we want an array of answers? */
     unsigned int wantarray : 1;
+
+    /* Has idic been built for this string? */
+
+    unsigned int idic_ok : 1;
 }
 text_fuzzy_t;
 
@@ -395,6 +454,219 @@ text_fuzzy_t;
 
 #endif /* HEADER */
 
+int
+minimum (int a, int b)
+{
+    if (a > b) {
+	return b;
+    }
+    return a;
+}
+
+int
+maximum (int a, int b)
+{
+    if (a > b) {
+	return a;
+    }
+    return b;
+}
+
+/* Dictionaries for the transposition searches. */
+
+/* For the binary search of dic->items. */
+
+static int
+item_comp (const void * a, const void * b)
+{
+    idic_item_t * ai = (idic_item_t *) a;
+    idic_item_t * bi = (idic_item_t *) b;
+    return (int) (ai->key) - (int) (bi->key);
+}
+
+/* For sorting the elements of worduniq into key order. */
+
+static int
+uniqcomp (const void * a, const void * b)
+{
+    return (int) (* (int *) a) - (int) (* (int *) b);
+}
+
+// s/stack/dic/g;
+
+/* Set up the array of "dic" to be a list of items sorted by the
+   characters of "word1". */
+
+static int
+idic_init_dic (idic_t * dic, const int * word1, int len1)
+{
+    int i;
+    int prev;
+    int top;
+#ifdef STACKALLOCOK
+    int worduniq[len1];
+#else
+    int * worduniq;
+#endif
+
+#ifndef STACKALLOCOK
+    CALLOC (worduniq, len1, sizeof (int));
+#endif
+    /* Put word1 into a copy then sort the copy to get the unique
+       elements. */
+    for (i = 0; i < len1; i++) {
+	worduniq[i] = word1[i];
+    }
+    qsort (worduniq, len1, sizeof (int), uniqcomp);
+
+    /* Count uniques. */
+
+    dic->max = 0;
+    /* There should not be any zeros in word1. */
+    prev = 0;
+    for (i = 0; i < len1; i++) {
+	int key;
+
+	key = worduniq[i];
+	if (key == prev) {
+	    continue;
+	}
+	dic->max++;
+	prev = key;
+    }
+    /* Fill "items" with the keys. Values are all zero at this point. */
+
+    CALLOC (dic->items, dic->max, sizeof (idic_item_t));
+
+    top = 0;
+
+    /* There should not be any zeros in word1. */
+    prev = 0;
+    for (i = 0; i < len1; i++) {
+	int j;
+	int found;
+	int key;
+
+	key = worduniq[i];
+	if (key == prev) {
+	    continue;
+	}
+	dic->items[top].key = key;
+	dic->items[top].position = 0;
+	top++;
+	if (top > dic->max) {
+	    fprintf (stderr, "%s:%d: dic overflow %d > %d.\n",
+		     __FILE__, __LINE__, top, dic->max);
+	    return -1;
+	}
+	prev = key;
+    }
+#ifndef STACKALLOCOK
+    free (worduniq);
+#endif
+    /* Because we sorted worduniq above, dic->items is already
+       sorted into key order, and we don't need to sort it with
+       item_comp to use bsearch. */
+    return 0;
+}
+
+int
+idic_free_dic (idic_t * dic)
+{
+    free (dic->items);
+    dic->items = 0;
+    return 0;
+}
+
+static idic_item_t * 
+idic_dicfind (idic_t * dic, int key)
+{
+    idic_item_t s = {0};
+    s.key = key;
+    /* If the number of items in dic is very small, it might be smart
+       to use a linear search. */
+    return bsearch (& s, dic->items, dic->max,
+		    sizeof (idic_item_t), item_comp);
+}
+
+/* Look for the most recent example of the character "key" in the
+   dic of letters from "word1". If it is not there anywhere, return
+   0. */
+
+int
+idic_find (idic_t * dic, int key)
+{
+    idic_item_t * found;
+    found = idic_dicfind (dic, key);
+    if (! found) {
+	return 0;
+    }
+    return found->position;
+}
+
+/* Set the value of dic[key] to value. */
+
+int
+idic_set (idic_t * dic, int key, int position)
+{
+    idic_item_t * found;
+    found = idic_dicfind (dic, key);
+    if (! found) {
+	fprintf (stderr,
+		 "%s:%d: could not set element %d: not found.\n",
+		 __FILE__, __LINE__, key);
+	return -1;
+    }
+    found->position = position;
+    return 0;
+}
+
+/* Reset this dic for another transposition match. */
+
+int
+idic_reset (idic_t * dic)
+{
+    int i;
+    for (i = 0; i < dic->max; i++) {
+	dic->items[i].position = 0;
+    }
+    return 0;
+}
+
+int
+adic_reset_dic (adic_t * dic)
+{
+    int i;
+    for (i = 0; i < UINT8_MAX; i++) {
+	dic->dic[i] = 0;
+    }
+    return 0;
+}
+
+int
+adic_free_dic (adic_t * dic)
+{
+    return 0;
+}
+
+int
+adic_find (adic_t * dic, uint8_t key)
+{
+    return dic->dic[key];
+}
+
+int
+adic_set (adic_t * dic, uint8_t key, int position)
+{
+    dic->dic[key] = position;
+    return position;
+}
+
+#include "ed-trans-char.h"
+#include "ed-trans-int.h"
+#include "edit-distance-char.h"
+#include "edit-distance-int.h"
+
 /* The following calculations need to be done twice, first when
    creating the alphabet and second when looking up a new character in
    it. The macro saves us from exasperating bugs. */
@@ -593,7 +865,6 @@ static int ualphabet_miss (text_fuzzy_t * tf, text_fuzzy_string_t * b)
 
 text_fuzzy_status_t text_fuzzy_compare_single (text_fuzzy_t * tf)
 {
-
     /* The edit distance between "tf->search_term" and the
        truncated version of "tf->buf". */
 
@@ -660,6 +931,13 @@ text_fuzzy_status_t text_fuzzy_compare_single (text_fuzzy_t * tf)
 
 	if (tf->transpositions_ok) {
 	    MESSAGE ("Transpositions OK.\n");
+	    if (! tf->idic_ok) {
+		text_fuzzy_string_t * t = & tf->text;
+		idic_init_dic (& tf->idic, t->unicode, t->ulength);
+		tf->n_mallocs++;
+		tf->idic_ok = 1;
+	    }
+	    idic_reset (& tf->idic);
 	    d = distance_int_trans (tf);
 	}
 	else {
@@ -719,6 +997,7 @@ text_fuzzy_status_t text_fuzzy_compare_single (text_fuzzy_t * tf)
 	   algorithm for "unsigned char". */
 
 	if (tf->transpositions_ok) {
+	    adic_reset_dic (& tf->adic);
 	    d = distance_char_trans (tf);
 	}
 	else {
@@ -788,7 +1067,7 @@ text_fuzzy_status_t text_fuzzy_get_candidates (text_fuzzy_t * text_fuzzy,
 
     candidates = malloc (sizeof (int) * n_candidates);
     FAIL (! candidates, memory_failure);
-    text_fuzzy->n_mallocs+=1;
+    text_fuzzy->n_mallocs++;
 
     last = text_fuzzy->first.next;
     i = 0;
@@ -1079,6 +1358,12 @@ text_fuzzy_status_t text_fuzzy_free_memory (text_fuzzy_t * text_fuzzy)
 	free (text_fuzzy->ualphabet.alphabet);
 	text_fuzzy->n_mallocs--;
     }
+    if (text_fuzzy->idic_ok) {
+	idic_free_dic (& text_fuzzy->idic);
+	text_fuzzy->idic_ok = 0;
+	text_fuzzy->n_mallocs--;
+    }
+
     OK;
 }
 
